# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A rag-bag of utility functions that did not fit anywhere else.

.. autoclass:: DeepDict
   :members:

.. autofunction:: smart_open

.. autofunction:: yesno

.. autofunction:: expand_path

.. autoclass:: ChangeDir

.. autofunction:: cached_iterator

"""

# Can be removed starting with python3.11
from __future__ import annotations

import configparser
import contextlib
import functools
import importlib
import io
import logging
import os
import pkgutil
import re
import shutil
import sys
import typing
from collections import defaultdict

RE_BRACKETS = re.compile("{[^{]*}")
LOGGER = logging.getLogger(__name__)


def copy(source, destination):
    """Copy ``source`` to ``destination``, creating directories if necessary."""
    os.makedirs(os.path.dirname(destination), exist_ok=True)
    shutil.copy2(source, destination)


def partial_format(string, *args, **kwargs):
    """Format string, leaving unfilled fields untouched.

    >>> partial_format("To {verb} or {negation} to {verb}", negation="not")
    'To {verb} or not to {verb}'
    """
    dictionary = {field[1:-1]: field for field in re.findall(RE_BRACKETS, string)}

    dictionary.update(kwargs)
    return string.format(*args, **dictionary)


@contextlib.contextmanager
def smart_open(filename, mode="w", encoding="utf8"):
    """Open filename, standard output, or nothing.


    :param str filename: If ``filename`` is:

        * ``""`` (the empty string): return a fake file object,
            which is empty if file is open for reading,
            and can be written in if file is open for writing (but content is then discarded);
        * ``"-"`` (a dash):
            read from standard input, or write to standard output (depending on mode);
        * any other: open the given file.
    :param str mode: Same as the ``mode`` parameter of :func:`open`.
    :param str encoding: Same as the ``encoding`` parameter of :func:`open`.
    """
    if (not filename) or (not filename.strip()):
        with io.StringIO() as file:
            yield file
    elif filename.strip() == "-":
        if "w" in mode:
            yield sys.stdout
        elif "r" in mode:
            yield sys.stdin
    else:
        with open(filename, mode, encoding=encoding) as file:
            yield file


def yesno(arg: typing.Union[bool, str, int, None]) -> bool:
    """Interpret some (mostly :class:`str`) variable as a boolean.

    >>> yesno("y")
    True
    >>> yesno("0")
    False
    >>> yesno("1")
    True
    >>> yesno("Yes")
    True
    >>> yesno("True")
    True
    >>> yesno("something senseless")
    False
    >>> yesno(None)
    False
    """
    if isinstance(arg, bool):
        return arg
    try:
        return bool(int(arg))
    except (ValueError, TypeError):
        pass
    if isinstance(arg, str):
        return arg.lower() in ["y", "yes", "true"]
    return bool(arg)


class ChangeDir(contextlib.AbstractContextManager):
    """Context manager to change and restore current directory."""

    # pylint: disable=too-few-public-methods

    def __init__(self, directory):
        if not directory:
            directory = "."
        self._newdir = directory
        self._olddir = os.getcwd()

    def __enter__(self, *args, **kwargs):
        self._olddir = os.getcwd()
        os.chdir(self._newdir)
        return super().__enter__(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        os.chdir(self._olddir)
        return super().__exit__(*args, **kwargs)


def expand_path(path):
    """Return ``path`` where environment variables and user directory have been expanded."""
    return os.path.expanduser(os.path.expandvars(path))


class DeepDict(defaultdict):
    """Dictionary of dictionary of ... of dictionaries.

    All the dictionaries (expeted the last one) are :class:`collections.defaultdict` objects.
    """

    def __init__(self, depth, dictionary=None):
        if dictionary is None:
            dictionary = {}

        self.depth = depth
        if depth == 1:
            factory = lambda: None
        else:
            factory = functools.partial(self.__class__, depth - 1)

        super().__init__(factory)
        for key, value in dictionary.items():
            self[key].update(value)

    @classmethod
    def from_configparser(cls, config: configparser.ConfigParser) -> DeepDict:
        """Create a :class:`DeepDict` object from a :class:`configparser.ConfigParser` object."""
        dictionary = cls(2)
        for section in config.sections():
            for option in config.options(section):
                dictionary[section][option] = config.get(section, option)
        return dictionary

    def get_subkey(self, subkey):
        """Return the first ``self[ANY][subkey]``, when ``ANY`` is any dictionary key."""
        for key in sorted(self):
            if subkey in self[key]:
                return self[key][subkey]
        raise KeyError(subkey)

    @property
    def __dict__(self):
        """Convert self to regular :type:`dict`."""
        if self.depth == 1:
            return dict(self)
        return {key: vars(value) for (key, value) in self.items()}

    def fill_blanks(self, other: dict):
        """Recursively copy values of ``other`` into ``self``.

        The values are copied only if those of ``self`` are not defined.
        """
        if self.depth == 1:
            for key in other:
                if key not in self:
                    self[key] = other[key]
        else:
            for key in other:
                self[key].fill_blanks(other[key])

    def copy(self):
        """Return a copy of ``self``."""
        new = self.__class__(depth=self.depth)
        if self.depth == 1:
            for key in self:
                new[key] = self[key]
        else:
            for key in self:
                new[key] = self[key].copy()
        return new


def iter_modules(path, prefix):
    """Iterate over modules.

    :arg list path: List of paths to look for modules in.
    :arg str prefix: String to output on the front of every module name on output.
    """
    for module_finder, name, __is_pkg in pkgutil.walk_packages(path, prefix):
        if name in sys.modules:
            module = sys.modules[name]
        else:
            try:
                # I don't fully understand the following lines.
                # They are copied from the Python documentation:
                # https://docs.python.org/3.10/library/importlib.html#importing-a-source-file-directly
                spec = module_finder.find_spec(name)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                sys.modules[name] = module
            except ImportError as error:
                LOGGER.debug(f"[plugins] Could not load module {name}: {error}")
                continue
            except Exception as error:  #  pylint: disable=broad-except
                LOGGER.debug(f"[plugins] Ignoring module {name}: {error}")
                continue
        yield module


def read_config(filename, *args, **kwargs):
    """Open filename, and parse it as a :module:`configparser` configuration file."""
    config = configparser.ConfigParser(*args, **kwargs)
    with open(filename, encoding="utf8") as file:
        config.read_file(file, source=filename)
    return config


def cached_iterator(func):
    """Like :func:`functools.cache`, but for iterators.

    That is, the first time the function is run, the returned iterable is stored (as a tuple),
    and next calls to the function return this tuple.
    """

    @functools.wraps(func)
    @functools.lru_cache()  # Can be replaced by @functools.cache once python3.8 is obsolete
    def wrapped(*args, **kwargs):
        return tuple(func(*args, **kwargs))

    return wrapped



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implement hook mechanism.

See :ref:`hooks` for more information.

.. note::

   This implementation of hooks rely on other parts of Évariste
   (:mod:`~evariste.plugins` for example),
   and cannot be used separatedly.

Example
-------

.. code-block:: python
   :caption: Example of hook mechanism

   import contextlib

   from evariste import hooks

   class A:

       @hooks.setmethodhook()
       def a(self):
           print("Running A.a()…")

   class B:

       @hooks.contexthook("A.a"):
       @contextlib.contextmanager
       def b(self):
           print("Before running A.a()…").
           yield
           print("After running A.a()…").

   # Let's go!
   A.a()

In this example, the ``A.a()`` method has been marked as accepting hooks,
and the ``B.b()`` method has been registered as a hook for ``A.a()``.

When ``A.a()`` is run (last line of the example),
although ``B`` has not been called directly,
``B.b()`` is called as well, as a registered hook. The output of this example is::

    Before running A.a()…
    Running A.a()…
    After running A.a()…

Get functions registered as hooks
---------------------------------

Method hooks
------------

Methods can be marked to accept hooks using the following function.

.. autofunction:: setmethodhook

Context hooks
-------------

Context hooks cannot be directly defined: every method hook is also a context hook.

Iteration hooks
---------------

Iteration hooks can be executed using :meth:`~evariste.plugins.Loader.applyiterhook`.

Register functions as hooks
---------------------------

.. autofunction:: hook

.. autofunction:: contexthook

.. autofunction:: methodhook

.. autofunction:: iterhook

"""

import collections
import functools
import typing


def hook(hooktype: str, name: str) -> typing.Callable:
    """Decorator to register a function or method as a hook.

    :param str hooktype:
        Type of hook (``"methodhook"`` or ``"contexthook"``, or whatever string you want).
    :param str name:
        Name of the target hook, of the form ``Class.methodname``
        (or ``Class`` only for the ``__init__`` method).
    """

    def wrapper(function):
        @functools.wraps(function)
        def wrapped(*args, **kwargs):
            return function(*args, **kwargs)

        hooked = getattr(function, "hooked", collections.defaultdict(set))
        hooked[hooktype].add(name)
        setattr(wrapped, "hooked", hooked)
        return wrapped

    return wrapper


def methodhook(name: str) -> typing.Callable:
    """Decorator to register a function or method as a method hook.

    For any string ``name``, ``methodhook(name)`` is a shortcut for ``hook("methodhook", name)``.
    """
    return hook("methodhook", name)


def contexthook(name: str) -> typing.Callable:
    """Decorator to register a function or method as a context hook.

    For any string ``name``, ``contexthook(name)`` is a shortcut for ``hook("contexthook", name)``.
    """
    return hook("contexthook", name)


def iterhook(name: str) -> typing.Callable:
    """Decorator to register a function or method as an iter hook.

    For any string ``name``, ``iterhook(name)`` is a shortcut for ``hook("iterhook", name)``.
    """
    return hook("iterhook", name)


def iter_hooks(instance):
    """Iterates over the method of `instance` registered as hooks."""
    for attrname in dir(instance):
        attr = getattr(instance, attrname)
        if callable(attr) and hasattr(attr, "hooked"):
            for hooktype, hooks in getattr(attr, "hooked").items():
                for item in hooks:
                    yield hooktype, item, attr


def setmethodhook(
    *, getter: typing.Union[None, typing.Callable] = None
) -> typing.Callable:
    """Decorator to mark that a method can accept method and context :ref:`hooks`.

    :param function getter:
        Function that, given the instance object as argument,
        returns a :class:`plugins.Loader` object.
        If ``None``, the default ``self.shared.builder.plugins`` is used
        (``self`` is supposed to have this attribute).
    """

    def decorator(function):
        """Actual decorator."""
        if function.__name__ == "__init__":
            hookname = function.__qualname__[: -len(function.__name__) - 1]
        else:
            hookname = function.__qualname__

        @functools.wraps(function)
        def wrapped(*args, **kwargs):
            """Wrapped function."""
            self = args[0]
            if getter is None:
                plugins = self.shared.builder.plugins
            else:
                plugins = getter(*args, **kwargs)

            return plugins.applymethodhook(hookname, function, *args, **kwargs)

        return wrapped

    return decorator



# Copyright Louis Paternault 2020
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the shared plugin"""

import unittest

from evariste import utils


class TestDeepDict(unittest.TestCase):
    """Testing :class:`evariste.utils.DeepDict`"""

    # pylint: disable=too-few-public-methods, no-member

    def test_fill_blanks(self):
        """Test that `fill_blanks` method works."""
        one = utils.DeepDict(2)
        one["a"]["a"] = "aa"
        one["a"]["b"] = "ab"
        one["b"]["a"] = "ba"

        two = utils.DeepDict(2)
        two["c"]["a"] = "ca"
        two["a"]["a"] = "WRONG"
        two["b"]["b"] = "bb"

        one.fill_blanks(two)

        self.assertEqual(one["a"]["a"], "aa")
        self.assertEqual(one["a"]["b"], "ab")
        self.assertEqual(one["b"]["a"], "ba")
        self.assertEqual(one["c"]["a"], "ca")
        self.assertEqual(one["b"]["b"], "bb")



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Parse setup file."""

import collections
import configparser
import os
import pathlib
import pprint
import shlex

from . import utils
from .errors import EvaristeError


def Options(*args, **kwargs):
    """Return a :class:`collections.defaultdict` object, default being `None`."""
    # pylint: disable=invalid-name
    return collections.defaultdict(lambda: None, *args, **kwargs)


def Sections(dictionary=None):
    """Return a :class:`collections.defaultdict` object, default being `Options`"""
    # pylint: disable=invalid-name
    if dictionary is None:
        dictionary = {}
    return collections.defaultdict(
        Options, {key: Options(value) for (key, value) in dictionary.items()}
    )


def string2config(string):
    """Parse ``string`` as the content of a configuration file."""
    config = configparser.ConfigParser(allow_no_value=True)
    config.read_string(string)
    return config


class Setup:
    """Representation ef Evariste setup.

    Precondition: If this object is built from a file,
    current working directory is the directory of this file.
    """

    def __init__(self, dictionary=None, *, filename=None):
        if dictionary is None:
            dictionary = {}
        self.dict = Sections(dictionary)

        # Load inherited setup files
        if filename is None:
            self._loaded = []
        else:
            self._loaded = [pathlib.Path(filename).resolve()]

        self._inherit()

        if self["setup"]["cachedir"] is None and filename is not None:
            self["setup"]["cachedir"] = f".{os.path.basename(filename)}.cache"

    def _inherit(self):
        for extends in shlex.split(self.get("setup", {}).get("extends", "")):
            # Inherit setup file
            absextends = pathlib.Path(extends).resolve()

            if not absextends.exists():
                # Raise error on non-existent files
                raise FileNotFoundError(extends)

            if absextends in self._loaded:
                # Prevent extends loop
                continue

            # Remove "setup.extends" option, so that it can be set by inherited setup
            del self["setup"]["extends"]

            self._loaded.append(absextends)
            with utils.ChangeDir(absextends.parent.as_posix()):
                with open(absextends.name, encoding="utf8") as file:
                    self.fill_blanks(string2config(file.read()))
                    self._inherit()

    def __iter__(self):
        yield from self.dict

    def __dict__(self):
        return self.dict

    def __getitem__(self, value):
        return self.dict[value]

    def get(self, key, default=None):
        """Alias to `self.dict.get`."""
        return self.dict.get(key, default)

    def __str__(self):
        # pylint: disable=consider-using-f-string
        return "{{{}}}".format(
            ", ".join([f"{key}: {value}" for key, value in self.dict.items()])
        )

    def __eq__(self, other):
        for section in set(self.keys()) | set(other.keys()):
            for option in set(self[section].keys()) | set(other[section].keys()):
                if self[section][option] != other[section][option]:
                    return False
        return True

    def fill_blanks(self, dictionary):
        """Fill unset self options with argument.

        Returns ``self`` for convenience.
        """
        for section in dictionary:
            if section in self.dict:
                for option in dictionary[section]:
                    if option not in self.dict[section]:
                        self.dict[section][option] = dictionary[section][option]
            else:
                self.dict[section] = Options(dictionary[section])
        return self

    def update(self, other, *, extend_list=False):
        """Update setup, with another setup or dict object.

        Similar to :meth:`dict.update`.

        :param boolean extend_list: If `True` and both values are lists,
            extend the first one with the second one.

        Return `self` for convenience.
        """
        for section in other:
            for option in other[section]:
                if (
                    extend_list
                    and isinstance(self.dict[section][option], list)
                    and isinstance(other[section][option], list)
                ):
                    self.dict[section][option].extend(other[section][option])
                else:
                    self.dict[section][option] = other[section][option]
        return self

    @classmethod
    def from_file(cls, filename):
        """Parse configuration file ``filename``."""
        with open(filename, encoding="utf8") as file:
            return cls(string2config(file.read()), filename=filename)

    def keys(self):
        """Return a new view of the setup’s keys"""
        return self.dict.keys()

    def pprint(self):
        """Pretty print of the object."""
        pprint.pprint(self.dict)

    @property
    def __dict__(self):
        """Return self, as a :class:`dict` of :class:`dict`."""
        dictionary = {}
        for section in self:
            dictionary[section] = dict(self[section])
        return dictionary

    def copy(self):
        """Return a copy of this object."""
        return self.__class__(vars(self))

    def items(self):
        """Iterate over (key, value) pairs, as 2-tuples."""
        yield from self.dict.items()


class SetupError(EvaristeError):
    """Error in setup file."""



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin to copy files at the end of compilation and rendering."""

import contextlib
import glob
import logging
import shutil

from ..hooks import contexthook
from . import Plugin

LOGGER = logging.getLogger("evariste.plugins.copy")


class Copy(Plugin):
    """Copy files at the end of compilation."""

    # pylint: disable=too-few-public-methods

    keyword = "copy"

    @contexthook("Builder.close")
    @contextlib.contextmanager
    def copy(self, builder):
        """Copy files (as configured in the configuration file) at the end of compilation."""
        # pylint: disable=unused-argument

        yield

        for key in self.local.setup:
            if key.startswith("copy"):
                if isinstance(self.local.setup[key], str):
                    arguments = self.local.setup[key].split()
                else:
                    arguments = self.local.setup[key]
                if len(arguments) <= 1:
                    raise ValueError(
                        f"""[copy] Option "{key}" should be "SOURCE [SOURCE] DEST"."""
                    )
                try:
                    sources = arguments[0:-1]
                    dest = arguments[-1]
                except (ValueError, TypeError) as error:
                    raise ValueError(
                        f"""[copy] Option "{key}" should be "SOURCE [SOURCE] DEST"."""
                    ) from error
                LOGGER.info(f"""Copying '{" ".join(sources)}' to '{dest}'.""")
                for source in sources:
                    for path in glob.iglob(source):
                        shutil.copytree(path, dest, dirs_exist_ok=True)



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Render tree as an HTML (body) page.

.. autoclass:: HTMLRenderer
   :members:

"""

import contextlib
import datetime
import logging

from .... import VERSION
from ..jinja2 import Jinja2Renderer
from . import file


class HTMLRenderer(Jinja2Renderer):
    """Render tree as an HTML div (without the ``<div>`` tags).

    The default template name is ``tree.html``,
    and such a default template is found in one of the template directories.
    """

    # pylint: disable=too-few-public-methods

    keyword = "renderer.html"

    #: Default template.
    #: This can be overloaded in the setup file.
    #: The template is looked for in any of the :ref:`templatedirs <jinja2_options>`.
    template: str = "tree.html"
    default_setup = {"href_prefix": "", "destfile": "index.html"}
    depends = ["renderer.html.readme.html", "renderer.html.file.default"]
    default_templatevar = {
        # pylint: disable=line-too-long, consider-using-f-string
        "aftertree": 'Generated using <a href="http://framagit.org/spalax/evariste">Évariste</a> version {}, on {}.'.format(
            VERSION, datetime.datetime.now().strftime("%c")
        )
    }



# Copyright Louis Paternault 2020-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of `evs compile`"""

import difflib
import filecmp
import glob
import os
import shutil
import subprocess
import sys
import unittest

from evariste import utils
from evariste.builder import Builder
from evariste.setup import Setup

SETUP = "evariste.setup"


class DynamicTest(type):
    """Metaclass that creates on-the-fly test methods.

    It creates a test method for each directory in `data`.
    """

    def __init__(cls, name, bases, nmspc):
        # pylint: disable=no-value-for-parameter
        super().__init__(name, bases, nmspc)
        for methodname, testmethod in cls._iter_testmethods():
            setattr(cls, methodname, testmethod)

    def _iter_testmethods(cls):
        """Iterate over dynamically generated test methods."""
        raise NotImplementedError()


class TestCompilation(unittest.TestCase, metaclass=DynamicTest):
    """Test that `evs compile` works, by comparing created files."""

    @classmethod
    def _iter_testmethods(cls):
        """Iterate over dynamically generated test methods"""
        for source in sorted(
            glob.glob(
                os.path.join(
                    os.path.dirname(__file__),
                    "data",
                    "*",
                )
            )
        ):
            if not os.path.exists(cls._setupfile(source)):
                continue
            yield (
                f"test_compile_{os.path.basename(source)}",
                cls._create_compile_test(source),
            )

    @staticmethod
    def _setupfile(source):
        return os.path.join(source, SETUP)

    @classmethod
    def _create_compile_test(cls, source):
        """Create and return the test method for test directory `source`."""

        def test_compile(self):
            """Test `source` directory.

            That is, in `source`:
            - run `evs compile evariste.setup`;
            - check that created directory `dest` is equal to `expected`.
            """
            with utils.ChangeDir(source):
                # Remove old destination directory and cache
                shutil.rmtree("dest", ignore_errors=True)
                shutil.rmtree(f".{SETUP}.cache", ignore_errors=True)
                os.makedirs("dest", exist_ok=True)

                # Run compilation
                with Builder.from_setupname(SETUP) as builder:
                    builder.compile()  # pylint: disable=no-member

                # Compare expected output to actual output
                self.assertTreeEqual("dest", "expected")

        with open(os.path.join(source, "description.txt"), encoding="utf8") as desc:
            test_compile.__doc__ = desc.read()

        return test_compile

    @staticmethod
    def _walk(directory):
        for root, __, files in os.walk(directory):
            for file in files:
                yield os.path.join(root, file)[len(directory) + 1 :]

    def assertTreeEqual(self, left, right):  # pylint: disable=invalid-name
        """Assert that directories `left` and `right` are equal.

        That is:
        - they contain the same files;
        - files are equal.

        Empty directories are ignored.
        """
        leftcontent = set(self._walk(left))
        rightcontent = set(self._walk(right))

        if leftcontent - rightcontent:
            filecmp.dircmp(left, right).report_full_closure()
            raise AssertionError(
                "Some files if first directory are missing in second directory."
            )
        if rightcontent - leftcontent:
            filecmp.dircmp(left, right).report_full_closure()
            raise AssertionError(
                "Some files in second directory are missing in first directory."
            )

        # If we got to this point, then leftcontent == rightcontent
        for name in leftcontent:
            if not filecmp.cmp(
                os.path.join(left, name), os.path.join(right, name), shallow=False
            ):
                filecmp.dircmp(left, right).report_full_closure()
                leftname = os.path.join(left, name)
                rightname = os.path.join(right, name)
                with open(leftname, encoding="utf8") as leftfile, open(
                    rightname, encoding="utf8"
                ) as rightfile:
                    sys.stdout.writelines(
                        difflib.unified_diff(
                            leftfile.readlines(),
                            rightfile.readlines(),
                            fromfile=leftname,
                            tofile=rightname,
                        )
                    )
                raise AssertionError(f"""File "{name}" differ in directories.""")



# Copyright Louis Paternault 2015-2021
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Actions that does nothing; simply provides a link to download file."""

import logging

from . import Action, Report

LOGGER = logging.getLogger(__name__)


class Raw(Action):
    """Do nothing. Source is not compiled."""

    keyword = "action.raw"
    priority = -float("inf")

    def match(self, value):
        # pylint: disable=unused-argument
        return True

    def compile(self, path):
        return Report(path, targets=[], success=True)


class NoPlugin(Action):
    """Special action used when no plugin is found."""

    keyword = "action.noplugin"
    priority = -float("inf")

    def match(self, value):
        # pylint: disable=unused-argument
        return False

    def compile(self, path):
        if "plugin" in path.config["action"]:
            # pylint: disable=line-too-long
            message = f"Error: Action plugin '{path.config['action']['plugin']}' not found for '{path.from_source}'."
            if (
                f"action.{path.config['action']['plugin']}"
                in self.shared.builder.plugins.available["action"]
            ):
                message += " Did you forget to enable it?"
        else:
            message = f"Error: No action plugin found for '{path.from_source}'."
        LOGGER.warning(message)
        return Report(
            path,
            success=False,
            log=message,
        )



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Access to VCS (git, etc.) versionned files.

Every path processed here is a :class:`pathlib.Path` object.

.. autoclass:: VCS
   :members:
   :special-members: __contains__

"""

import abc
import os
import pathlib
import typing
from datetime import datetime

from ...errors import EvaristeError
from ...tree import Root
from .. import Plugin


class NoRepositoryError(EvaristeError):
    """No repository contains the given path."""

    def __init__(self, vcstype, directory):
        super().__init__()
        self.directory = directory
        self.vcstype = vcstype

    def __str__(self):
        # pylint: disable=line-too-long
        return f"Could not find any {self.vcstype} repository containing directory '{self.directory}'."


class VCS(Plugin, metaclass=abc.ABCMeta):
    """Generic class to access to versionned files.

    To write a new VCS plugin, one has to subclass this class, and implement every abstract method
    (see for instance the implementation of :class:`evariste.plugin.vcs.git.Git`).
    """

    plugin_type = "vcs"
    global_default_setup = {"setup": {"source": "."}}

    @property
    def source(self) -> pathlib.Path:
        """Return an absolute version of source setup option."""
        return pathlib.Path(self.shared.setup["setup"]["source"]).resolve()

    @abc.abstractmethod
    def walk(self) -> typing.Iterable[pathlib.Path]:
        """Iterate versionned files, descendants of source (as defined by setup file)."""
        raise NotImplementedError()

    @abc.abstractmethod
    def __contains__(self, path: pathlib.Path) -> bool:
        """Return ``True`` iff ``path`` is versionned."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def workdir(self) -> pathlib.Path:
        """Return path of the root of the repository."""
        raise NotImplementedError()

    def from_repo(self, path: pathlib.Path) -> pathlib.Path:
        """Return ``path``, relative to the repository root."""
        return path.resolve().relative_to(self.workdir)

    def last_modified(self, path: pathlib.Path) -> datetime:
        """Return the datetime of last modification."""
        # pylint: disable=no-self-use
        return datetime.fromtimestamp(os.path.getmtime(path.as_posix()))



# Copyright Louis Paternault 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of `evs compile`"""

import difflib
import filecmp
import glob
import os
import pathlib
import shutil
import subprocess
import sys
import unittest

from evariste import utils
from evariste.builder import Builder

SETUP = {
    "setup": {
        "vcs": "vcs.git",
        "source": pathlib.Path(__file__).parent / "data" / "tree",
        "plugins": ["vcs.git"],
    }
}


class TestIgnore(unittest.TestCase):
    """Test that path are correctly ignored."""

    def test_ignore(self):
        """Test that path are correctly ignored."""
        with utils.ChangeDir(pathlib.Path(__file__).parent / "data"):
            with Builder.from_setupdict(SETUP) as builder:
                builder.compile()  # pylint: disable=no-member
            self.assertEqual(
                set(
                    str(path.from_source)
                    for path in builder.tree.walk()  # pylint: disable=no-member
                ),
                set(
                    (
                        "a/b/tagada.md",
                        "baz",
                        "c/tata.txt",
                        "c/titi.txt",
                        "c/toto.txt",
                        "d/d/d/d/d/tata.txt",
                        "d/d/d/d/d/titi.txt",
                        "d/d/d/d/d/toto.txt",
                        "d/d/d/d/titi.txt",
                        "d/d/d/d/toto.txt",
                        "d/d/d/tata.txt",
                        "d/d/d/toto.txt",
                        "d/d/tata.txt",
                        "d/d/titi.txt",
                    )
                ),
            )



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the setup plugin"""

import os
import unittest

from evariste.plugins import Plugin
from evariste.setup import Setup
from evariste.shared import Shared
from evariste.utils import ChangeDir

################################################################################

DATADIR = os.path.join(os.path.dirname(__file__), "data")


class First(Plugin):
    """Test plugin"""

    # pylint: disable=abstract-method, too-few-public-methods

    plugin_type = "test"
    default_setup = {"foo": 1, "bar": 2}


class Second(First):
    """Another test plugin"""

    # pylint: disable=abstract-method, too-few-public-methods

    keyword = "test.second"
    default_setup = {"foo": 3, "baz": 4}


class TestDefaultSetup(unittest.TestCase):
    """Testing default setup"""

    # pylint: disable=too-few-public-methods

    def test_inheritance(self):
        """Testing default setup inheritance"""
        plugin = Second(
            Shared(setup=Setup({"setup": {"vcs": "vcs.none"}}), builder=None)
        )
        self.assertEqual(plugin.shared.setup["test.second"]["foo"], 3)
        self.assertEqual(plugin.shared.setup["test.second"]["bar"], 2)
        self.assertEqual(plugin.shared.setup["test.second"]["baz"], 4)

        self.assertEqual(plugin.local.setup["foo"], 3)
        self.assertEqual(plugin.local.setup["bar"], 2)
        self.assertEqual(plugin.local.setup["baz"], 4)


################################################################################


class TestEqual(unittest.TestCase):
    """Test equality method of setup objects."""

    def test_empty(self):
        """Test the :meth:`Setup.__eq__` method on empty setup objects."""
        self.assertEqual(Setup({}), Setup({}))

    def test_same_dict(self):
        """Test the :meth:`Setup.__eq__` method on object built with the same dictionary."""
        base = {"first": {"one": 11, "two": 12}, "second": {"one": 21, "two": 22}}
        self.assertEqual(Setup(base), Setup(base))

    def test_default_objects(self):
        """Test the :meth:`Setup.__eq__` method on object having default values."""
        setup1 = Setup(
            {"first": {"one": 11, "two": 12}, "second": {"one": 21, "two": 22}}
        )
        setup2 = Setup(vars(setup1))

        # Creating default objects
        setup1["three"]  # pylint: disable=pointless-statement
        setup1["one"]["three"]  # pylint: disable=pointless-statement
        setup1["four"]["one"]  # pylint: disable=pointless-statement

        self.assertEqual(setup1, setup2)

    def test_different(self):
        """Test the :meth:`Setup.__eq__` method on different objects."""
        setup1 = Setup(
            {"first": {"one": 11, "two": 12}, "second": {"one": 21, "two": 22}}
        )

        with self.subTest():
            setup2 = setup1.copy()
            setup2["first"]["one"] = "foo"
            self.assertNotEqual(setup1, setup2)

        with self.subTest():
            setup3 = setup1.copy()
            setup3["first"]["three"] = "foo"
            self.assertNotEqual(setup1, setup3)

        with self.subTest():
            setup4 = setup1.copy()
            setup4["three"]["one"] = 31
            self.assertNotEqual(setup1, setup4)


################################################################################


class TestUpdate(unittest.TestCase):
    """Testing method :meth:`setup.Setup.update`."""

    def test_update(self):
        """Test the :meth:`Setup.update` method."""
        setup = Setup(
            {"first": {"one": 11, "two": 12}, "second": {"one": 21, "two": 22}}
        )
        setup.update({"first": {"one": "UPDATED", "three": 13}, "three": {"one": 31}})
        self.assertEqual(
            setup,
            Setup(
                {
                    "first": {"one": "UPDATED", "two": 12, "three": 13},
                    "second": {"one": 21, "two": 22},
                    "three": {"one": 31},
                }
            ),
        )


################################################################################


class TestExtends(unittest.TestCase):
    """Test inheritance of setup files."""

    def test_cycle(self):
        """Test that cycles are detected."""
        with ChangeDir(os.path.join(DATADIR, "cycle")):
            setup = Setup.from_file("a.setup")
            self.assertEqual(setup["setup"]["aaa"], "aaa")
            self.assertEqual(setup["setup"]["bbb"], "bbb")
            self.assertEqual(setup["setup"]["ccc"], "ccc")

    def test_absent(self):
        """Test that non-existent "extends = FOO" setup file raises an error."""
        with ChangeDir(os.path.join(DATADIR, "empty")):
            with self.assertRaises(FileNotFoundError):
                Setup({"setup": {"extends": "does/not/exists"}})



# Copyright Louis Paternault 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of `evs compile`"""

import difflib
import filecmp
import glob
import os
import pathlib
import shutil
import subprocess
import sys
import unittest

from evariste import utils
from evariste.builder import Builder

SETUP = {
    "setup": {
        "vcs": "vcs.git",
        "source": pathlib.Path(__file__).parent / "data",
        "plugins": ["vcs.git"],
    }
}


def get(obj, *keys):
    """Return `obj[keys[0]][keys[1]][…]`, or `None` if there is a KeyError.

    >>> ddict = {"a": {"aa": 1}, "b": {"bb": {"bbb": 2}}}
    >>> print(get(ddict, "a", "aa"))
    1
    >>> print(get(ddict, "b", "bb", "bbb"))
    2
    >>> print(get(ddict, "a", "c"))
    None
    >>> print(get(ddict, "b", "bb", "c"))
    None
    """
    if keys:
        try:
            return get(obj[keys[0]], *keys[1:])
        except KeyError:
            return None
    return obj


class TestRecursive(unittest.TestCase):
    """Test setup.source and setup.recursive options."""

    def test_recursive(self):
        """Test setup.source and setup.recursive options."""
        with utils.ChangeDir(pathlib.Path(__file__).parent / "data"):
            with Builder.from_setupdict(SETUP) as builder:
                builder.compile()  # pylint: disable=no-member
            self.assertEqual(
                {
                    str(path.from_source): path.config["foo"]["bar"]
                    for path in builder.tree.walk()  # pylint: disable=no-member
                },
                {
                    "a/aa/aaa": "a",
                    "b/bb/bbb": "bb",
                    "c/cc": None,
                    "d/dd/ddd": "d",
                    "e/ee": "e-recursive",
                    "f/ff/fff": "f",
                },
            )



# Copyright Louis Paternault 2015-2021
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Render images as HTML"""

from . import HtmlFileRenderer

IMAGE_EXTENSIONS = ["png", "jpg", "jpeg", "gif"]


class HtmlImageRenderer(HtmlFileRenderer):
    """Render images as HTML"""

    priority = 0
    keyword = "renderer.html.file.image"
    template = "image"

    def match(self, filename):
        return filename.as_posix().split(".")[-1].lower() in IMAGE_EXTENSIONS



# Copyright Louis Paternault 2016-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Text renderer: print report on terminal."""

import contextlib
import logging
import sys

import unidecode

from ...hooks import contexthook
from ...utils import yesno
from .. import Plugin

LOGGER = logging.getLogger("evariste")


class TreePrettyPrinter:
    """Pretty print a tree"""

    # pylint: disable=too-few-public-methods

    # This class could be a function, but it is a recursive function, and
    # turning it into a class allows to tell appart arguments that change
    # for every recursive call, and arguments that never change.

    def __init__(
        self,
        *,
        print_function=print,
        report=False,
        color=False,
        reverse=False,
        display=None,
    ):
        self.print_function = print_function
        self.report = report
        self.color = color
        self.reverse = reverse
        self.display = display

    def pprint(self, tree, *, prefix="", last=True):
        """Pretty print the given tree"""
        # pylint: disable=too-many-branches
        line = []

        txttree = ""
        txttree += prefix
        if last:
            txttree += {False: "└", True: "┌"}[self.reverse]
        else:
            txttree += "├"
        txttree += "─"
        if tree.is_dir():
            txttree += {True: "┴", False: "┬"}[self.reverse]
        else:
            txttree += "─"
        txttree += "╼"
        line.append(txttree)

        if self.report:
            if tree.report.success:
                if self.color:
                    reportstr = "\033[92m ok \033[0m"
                else:
                    reportstr = " ok "
            else:
                if self.color and tree.is_dir():
                    reportstr = "\033[93mfail\033[0m"
                elif self.color and not tree.is_dir():
                    reportstr = "\033[91mfail\033[0m"
                else:
                    reportstr = "fail"
            reportstr = f"[{reportstr}]"
            line.append(reportstr)

        line.append(str(tree.relativename))

        if not self.reverse:
            self.print_function(" ".join(line))

        if tree.is_dir():

            if last:
                postfix = "  "
            else:
                postfix = "│ "

            subtree = [branch for branch in tree.values() if self.display(branch)]
            if self.reverse:
                lastcounter = 1
            else:
                lastcounter = len(subtree)
            counter = 0
            for sub in sorted(subtree, reverse=self.reverse):
                counter += 1
                self.pprint(sub, prefix=prefix + postfix, last=(counter == lastcounter))

        if self.reverse:
            self.print_function(" ".join(line))


ALLOWED_DISPLAY_OPTIONS = ["errors", "all", "errors_or_all"]


class TextRenderer(Plugin):
    """Text renderer: print report on terminal."""

    keyword = "renderer.text"
    default_setup = {
        "color": "auto",
        "ascii": False,
        "reverse": False,
        "display": "all",
    }

    @contexthook("Builder.compile")
    @contextlib.contextmanager
    def render(self, builder):
        """Print the report to the standard output."""

        yield

        # Option 'ascii'
        if yesno(self.local.setup["ascii"]):

            def asciiprint(text):
                """Convert argument to ascii, and print it."""
                print(unidecode.unidecode(text))

            print_function = asciiprint
        else:
            print_function = print

        # Option 'reverse'
        reverse = yesno(self.local.setup["reverse"])

        # Option 'display'
        displayoption = self.local.setup["display"]
        if displayoption not in ALLOWED_DISPLAY_OPTIONS:
            LOGGER.warning(
                (
                    # pylint: disable=consider-using-f-string
                    "[renderer.text] Option 'display' must be one of {} (currently {})."
                ).format(
                    ", ".join((f"'{option}'" for option in ALLOWED_DISPLAY_OPTIONS)),
                    f"'{displayoption}'",
                )
            )
            displayoption = self.default_setup["display"]
        if displayoption == "errors_or_all":
            if builder.tree.report.success:
                displayoption = "all"
            else:
                displayoption = "errors"
        if displayoption == "errors":
            display = lambda tree: not tree.report.success
        else:
            display = lambda tree: True

        # Option 'color'
        color = self.local.setup["color"]
        if color.lower() == "auto":
            color = sys.stdout.isatty()  # pylint: disable=no-member
        color = yesno(color)

        # Render
        TreePrettyPrinter(
            report=True,
            color=color,
            print_function=print_function,
            reverse=reverse,
            display=display,
        ).pprint(builder.tree)



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line options"""

import argparse
import logging
import os
import sys
import textwrap

from ... import VERSION

LOGGER = logging.getLogger("evariste")
LOGGING_LEVELS = {
    -1: 100,  # Quiet
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
    3: 1,
}


def _get_logging_level(verbose):
    """Turn command line verbosity into :mod:`logging` verbosity."""
    if verbose in LOGGING_LEVELS:
        return LOGGING_LEVELS[verbose]
    if verbose is None:
        return LOGGING_LEVELS[0]
    if verbose > max(LOGGING_LEVELS.keys()):
        return LOGGING_LEVELS[max(LOGGING_LEVELS.keys())]
    raise NotImplementedError()


def _setup_type(name):
    """Check the argument and return its value.

    The argument must be an existing file.
    """
    if not os.path.exists(name):
        raise argparse.ArgumentTypeError(f"File '{name}' does not exist.")
    if not os.path.isfile(name):
        raise argparse.ArgumentTypeError(f"File '{name}' is not a file.")
    return name


class Options(argparse.Namespace):
    """Namespace of command line options.

    Added ability to iterate over option names.
    """

    # pylint: disable=too-few-public-methods

    def __str__(self):
        return str(
            {
                (attr, getattr(self, attr))
                for attr in dir(self)
                if not attr.startswith("_")
            }
        )

    def __iter__(self):
        for attr in dir(self):
            if not attr.startswith("_"):
                yield attr


def commandline_parser():
    """Return a command line parser."""

    parser = argparse.ArgumentParser(
        prog="evariste",
        description=("Recursively compile files in a directory, and render result."),
        epilog=(
            "Note that `evariste ARGS` and `evs compile ARGS` are the same command."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--version",
        help="Show version",
        action="version",
        version="%(prog)s " + VERSION,
    )

    parser.add_argument(
        "-v", "--verbose", help="Verbose. Repeat for more details.", action="count"
    )

    parser.add_argument(
        "-q",
        "--quiet",
        help="Quiet. Does not print anything to standard output.",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-j",
        "--jobs",
        help=(
            "Specify the number of jobs to run simultaneously. Default is one "
            "more than the number of CPUs. "
        ),
        action="store",
        default=os.cpu_count() + 1,
        type=int,
    )

    parser.add_argument(
        "-B",
        "--always-compile",
        dest="always_compile",
        help="Unconditionally make all targets",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "setup",
        metavar="SETUP",
        help=textwrap.dedent(
            """
            Setup file to process.
            """
        ),
        type=_setup_type,
    )

    return parser


def get_options():
    """Return the namespace of (processed) command line options."""
    try:
        arguments = commandline_parser().parse_args(namespace=Options())

        # Process --verbose and --quiet
        # pylint: disable=no-member, attribute-defined-outside-init
        if arguments.quiet:
            arguments.verbose = -1
        LOGGER.setLevel(_get_logging_level(arguments.verbose))

        return arguments
    except argparse.ArgumentTypeError as error:
        sys.stderr.write(str(error))
        sys.stderr.write("\n")
        sys.exit(2)



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract utilities for file renderers using Jinja2.

.. autoclass:: Jinja2FileRenderer
   :members:

"""

# Can be removed starting with python3.11
from __future__ import annotations

import typing

import jinja2

from ... import Plugin

if typing.TYPE_CHECKING:
    from ...tree import Tree


class Jinja2FileRenderer(Plugin):
    """Renderer of file using jinja2.

    This is an abstract class that defines a default renderer for files.

    From within a template,
    the `macro <https://jinja.palletsprojects.com/en/3.0.x/templates/#macros>`__
    ``render_file`` can be called, which:

    - looks for the first plugin that matches this file
      (that is, the first plugin where :meth:`Jinja2FileRenderer.match` returns ``True``;
    - calls :meth:`Jinja2FileRenderer.render`, and returns its return value.

    To implement such a renderer, you can:

    - write a :file:`file/default` template that defines a ``file()`` macro;
    - set the :attr:`Jinja2FileRenderer.extension`,
      and write a :file:`file/default.{extension}` template, that defines a ``file()`` macro;
    - overwrite the :meth:`Jinja2FileRenderer.render` method,
      if the default implementation does not pleas you.

    You can also overwrite :meth:`Jinja2FileRenderer.match`,
    so that your subplugin cannot be applied to *any* file, but only to some of them.
    """

    keyword = None
    #: Extension that is automatically added at the end of the template name when searching them.
    extension: typing.Optional[str] = None
    priority = -float("inf")

    #: Name of the template rendering files.
    template: str = "default"

    def match(self, filename: Tree) -> bool:
        """This is the default renderer, that matches everything."""
        # pylint: disable=unused-argument, arguments-renamed
        return True

    def render(self, filename: Tree, context: jinja2.runtime.Context) -> str:
        """Render ``tree``, which is a :class:`~evariste.tree.File`

        By default, call the ``file()``
        `macro <https://jinja.palletsprojects.com/en/latest/templates/#macros>`__,
        with ``filename`` as argument, and returns its value.
        """
        # pylint: disable=arguments-differ
        return context[f"file_{self.template}"](filename)



# Copyright Louis Paternault 2021-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin log progress."""

import contextlib
import logging
import threading

from ..hooks import contexthook
from . import Plugin


class Logging(Plugin):
    """Log stuff"""

    # pylint: disable=too-few-public-methods

    keyword = "logging"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lock = threading.Lock()
        self.logger = logging.getLogger("evariste.tree")
        self.count = 0
        self.total = 0

    @contexthook("Builder.compile")
    @contextlib.contextmanager
    def compile_builder(
        self, builder, *args, **kwargs
    ):  # pylint: disable=unused-argument
        """Save the total number of files to compile."""
        self.total = builder.tree.count()
        yield

    @contexthook("File.compile")
    @contextlib.contextmanager
    def compile_file(self, tree):
        """Log file compilation (including progress)."""
        with self.lock:
            self.count += 1
        count = self.count
        self.logger.info(
            f"Compiling [{count: >{len(str(self.total))}}/{self.total}] "
            f"'{tree.from_source}'…"
        )

        yield

        if tree.report.success:
            self.logger.info(
                f"Compiling [{count: >{len(str(self.total))}}/{self.total}] "
                f"'{tree.from_source}': success."
            )
        else:
            self.logger.info(
                f"Compiling [{count: >{len(str(self.total))}}/{self.total}] "
                f"'{tree.from_source}': failed."
            )



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Plugin base class and plugin loader

Every plugin is a subclass of :class:`Plugin` (see :ref:`write` for more information).

The :class:`Loader` class finds and loads the plugins.

Constants
---------

.. autodata:: MANDATORY_PLUGINS

:class:`Plugin`
---------------

.. autoclass:: Plugin
   :members:

:class:`Loader`
---------------

.. autoclass:: Loader
   :members: get_plugin, iter, values, items, applyiterhook, match

Functions
---------

.. autofunction:: find_plugins

Exceptions
----------

.. autoclass:: NoMatch

"""

import collections
import contextlib
import functools
import logging
import os
import shlex
import sys
import typing

from .. import errors, hooks, setup, utils
from ..shared import Shared

LOGGER = logging.getLogger("evariste")

#: Set of mandatory plugins: plugins that are loaded by default, and cannot be disabled.
MANDATORY_PLUGINS = {
    "action.cached",
    "action.directory",
    "action.noplugin",
    "action.raw",
    "changed",
    "logging",
    "tree",
}


class NoMatch(errors.EvaristeError):
    """No plugin found matching ``value``."""

    def __init__(self, value, available):
        super().__init__()
        self.value = value
        self.available = available

    def __str__(self):
        return f"Value '{self.value}' does not match any of {self.available}."


class SameKeyword(errors.EvaristeError):
    """Two plugins have the same keyword."""

    def __init__(self, keyword, plugin1, plugin2):
        super().__init__()
        self.keyword = keyword
        self._plugins = (plugin1, plugin2)

    def __str__(self):
        # pylint: disable=line-too-long
        return f"""Plugins '{self._plugins[0].__name__}' and '{self._plugins[1].__name__}' have the same keyword '{self.keyword}'."""


class NotAPlugin(errors.EvaristeError):
    """Superclass of plugins is not a plugin."""

    def __init__(self, obj):
        super().__init__()
        self.obj = obj

    def __str__(self):
        return (
            """Class '{obj.__module__}.{obj.__name__}' is not a plugin """
            """(it should inherit from """
            """'{superclass.__module__}.{superclass.__name__}')."""
        ).format(obj=self.obj, superclass=Plugin)


class PluginNotFound(errors.EvaristeError):
    """Plugin cannot be found."""

    def __init__(self, keyword):
        super().__init__()
        self.keyword = keyword

    def __str__(self):
        return f"Cannot find plugin '{self.keyword}'."


@functools.total_ordering
class Plugin:
    """Plugin base: all imported plugins must be subclasses of this class.

    See :ref:`write` to see how to write a new plugin.

    :param Shared shared: The object :ref:`shared <shared>` among plugins.
    """

    # pylint: disable=too-few-public-methods

    #: Keyword plugin, used to reference it: it is used
    #: to :ref:`enable plugins in the setup file <plugins>`,
    #: to name its section in the :ref:`setup file <setup>`, etc.
    keyword: typing.Union[None, str] = None

    #: When Évariste has to choose *one* plugin among several one,
    #: it chooses the one with higher priority.
    priority: int = 0

    #: Default value for section ``self.keyword`` in the :ref:`setup file <setup>`.
    #: It may be overwritten by data provided by user in the :ref:`setup file <setup>`.
    #: See :ref:`default_setup`.
    default_setup: typing.Dict[str, str] = {}

    #: Default values for setup file. See :ref:`default_setup`.
    global_default_setup: typing.Dict[str, typing.Dict[str, str]] = {}

    #: Type of the plugin.
    #: Plugins of the same type gather some common behaviour.
    plugin_type: str = ""

    #: Set of plugins this plugin depends on (as a list of :class:`str`):
    #: when this plugin is enabled, those plugins are enabled as well.
    depends: typing.Set[str] = {}

    def __init__(self, shared):
        #: Common data shared with every :class:`Tree` and :class:`~evariste.plugins.Plugin`
        #: of this :class:`~evariste.builder.Builder`.
        self.shared: Shared = shared

        #: Same as :attr:`Plugin.shared`, but from this plugin point of view:
        #: see :meth:`evariste.shared.Shared.get_plugin_view` and :ref:`plugin_local`.
        self.local = shared.get_plugin_view(self.keyword)

        self._set_default_setup()

    def _set_default_setup(self):
        """Set default value for this plugin setup, if necessary."""
        default = setup.Setup()
        for parent in reversed(self.__class__.mro()):  # pylint: disable=no-member
            if hasattr(parent, "global_default_setup"):
                default.update(parent.global_default_setup)
            if hasattr(parent, "default_setup"):
                default.update({self.keyword: parent.default_setup})
        self.shared.setup.fill_blanks(default)

    def match(self, value, *args, **kwargs) -> bool:  # pylint: disable=unused-argument
        """Return ``True`` iff ``value`` matches ``self``.

        Default is keyword match.
        This method can be overloaded by subclasses.
        """
        return value == self.keyword

    def __lt__(self, other):
        priority = self.priority
        if callable(priority):
            priority = priority()  # pylint: disable=not-callable
        other_priority = other.priority
        if callable(other_priority):
            other_priority = other_priority()
        if priority == other_priority:
            return self.keyword < other.keyword
        return priority < other_priority


def find_plugins(
    libdirs: typing.Union[typing.Iterable[str], None] = None
) -> typing.Iterator[Plugin]:
    """Iterate over available plugins.

    :param typing.Iterable[str] libdirs:
        Additional iterable of directories where plugins can be found.
    """
    if libdirs is None:
        libdirs = []

    path = []
    path.extend(
        [
            os.path.join(utils.expand_path(item), "plugins")
            for item in [".evariste", "~/.local/evariste", "~/.evariste"]
        ]
    )
    path.extend([utils.expand_path(item) for item in libdirs])
    path.extend([os.path.join(item, "evariste") for item in sys.path])

    yielded = set()
    for module in utils.iter_modules(path, "evariste."):
        for attr in dir(module):
            if attr.startswith("_"):
                continue
            obj = getattr(module, attr)

            if isinstance(obj, type) and issubclass(obj, Plugin):
                if obj.keyword is None:
                    continue
                if obj in yielded:
                    continue
                yielded.add(obj)
                yield obj


def find_plugins_sorted(libdirs=None):
    """Like :fun:`find_plugins`, but returns a :class:`utils.DeepDict`.

    The returned object is a dictionary of dictionaries:
    - first keys are plugin types;
    - second keys are plugin keywords;
    - values are the only plugin of said type and said keyword.
    """
    plugindict = utils.DeepDict(2)

    for plugin in find_plugins(libdirs):
        if plugin.keyword in plugindict[plugin.plugin_type]:
            raise SameKeyword(
                plugin.keyword,
                plugin,
                plugindict[plugin.plugin_type][plugin.keyword],
            )

        plugindict[plugin.plugin_type][plugin.keyword] = plugin
    return plugindict


def get_libdirs(libdirs):
    """Convert `libdirs` setup option (as a string) to a list of path (as strings)."""
    if not libdirs:
        return []
    if isinstance(libdirs, str):
        return shlex.split(libdirs)
    if isinstance(libdirs, list):
        return libdirs
    return []


class Loader:
    """Load plugins

    :param evariste.shared.Shared shared: The :ref:`shared <shared>` object among plugins.

    The constructor (:meth:`Loader.__init__`):

    - reads the :ref:`setup file <setup>` (looking for the :ref:`libdirs <libdirs>` option);
    - search all plugins (subclasses of :class:`Plugin`);
    - instanciate:

      - the :const:`mandatory plugins <MANDATORY_PLUGINS>`,
      - those enabled in the :ref:`setup file <setup>`,
      - and their dependencies;

    - store them in some attribute, so that they can be accessed later.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, *, shared):
        # pylint: disable=too-many-branches
        self.shared = shared

        # Load plugins:
        # Given the avaialble plugins, only loads relevant plugins.
        to_load = set()
        self.available = find_plugins_sorted(
            get_libdirs(self.shared.setup["setup"]["libdirs"])
        )

        # Step 0: Convert setup options in a list of keyword
        if self.shared.setup["setup"]["plugins"] is None:
            self.shared.setup["setup"]["plugins"] = []
        else:
            if isinstance(self.shared.setup["setup"]["plugins"], str):
                self.shared.setup["setup"]["plugins"] = shlex.split(
                    self.shared.setup["setup"]["plugins"]
                )
            elif isinstance(self.shared.setup["setup"]["plugins"], list):
                pass
            else:
                raise ValueError(
                    (
                        # pylint: disable=consider-using-f-string, line-too-long
                        "'Setup[setup][plugins]' should be a string or a list (is now {}: '{}')."
                    ).format(
                        type(self.shared.setup["setup"]["plugins"]),
                        self.shared.setup["setup"]["plugins"],
                    )
                )

        # Step 1: Add enabled plugins
        for keyword in self.shared.setup["setup"]["plugins"]:
            try:
                self.available.get_subkey(keyword)
            except KeyError as error:
                raise PluginNotFound(keyword) from error
            to_load.add(keyword)

        # Step 2: Enable/disable plugins according to their "enable" configuration
        for types in self.available.values():
            for plugin in types.values():
                if "enable" in self.shared.setup[plugin.keyword]:
                    if utils.yesno(self.shared.setup[plugin.keyword]["enable"]):
                        to_load.add(plugin.keyword)
                    elif plugin.keyword in MANDATORY_PLUGINS:
                        LOGGER.warning(
                            f"""Attemp to disable mandatory plugin "{plugin.keyword}" ignored."""
                        )
                        continue
                    else:
                        to_load.discard(plugin.keyword)

        # Step 3: Add default plugins
        to_load |= MANDATORY_PLUGINS

        # Step 4: Manage dependencies
        to_process = to_load.copy()
        processed = set()
        while to_process:
            keyword = to_process.pop()

            try:
                # Check if plugin exists
                self.available.get_subkey(keyword)
            except KeyError as error:
                raise PluginNotFound(keyword) from error

            for dependency in self.available.get_subkey(keyword).depends:
                try:
                    # Check if plugin exists
                    self.available.get_subkey(dependency)
                except KeyError as error:
                    raise PluginNotFound(dependency) from error
                if (dependency not in to_load) and (dependency not in processed):
                    to_load.add(dependency)
                    to_process.add(dependency)
            processed.add(keyword)

        # Step 5: Instanciate plugins
        self._plugins = collections.defaultdict(dict)
        self._hooks = collections.defaultdict(lambda: collections.defaultdict(set))
        for keyword in to_load:
            plugin = self.available.get_subkey(keyword)(self.shared)
            self._plugins[plugin.plugin_type][keyword] = plugin
            for hooktype, name, function in hooks.iter_hooks(plugin):
                self._hooks[hooktype][name].add(function)

    def values(
        self, plugin_type: typing.Optional[str] = None
    ) -> typing.Iterable[Plugin]:
        """Iterate over plugins.

        :param typing.Optional[str] plugin_type: See :meth:`Loader.iter`.
        """
        if plugin_type is None:
            for ptype in self._plugins:
                yield from self._plugins[ptype].values()
        else:
            yield from self._plugins[plugin_type].values()

    def applymethodhook(
        self, hookname: str, function: typing.Callable, *args, **kwargs
    ):
        """Apply a :ref:`method hook <methodhook>`.

        That is, run stuff before and after a method call.

        :param str hookname: Name of the hook to apply.
        :param function function: Function to which the hook is to be applied.
        :param list args: List of parameters to pass to hooks and ``function``.
        :param dict kwargs: Dictionary of parameters to pass to hooks and ``function``.
        """
        with contextlib.ExitStack() as stack:
            for hook in sorted(self._hooks["contexthook"][hookname], key=str):
                stack.enter_context(hook(*args, **kwargs))
            for hook in sorted(self._hooks["methodhook"][hookname], key=str):
                function = hook(function)

            return function(*args, **kwargs)

    def applyiterhook(self, hookname: str, *args, **kwargs) -> typing.Iterable:
        """Apply an :ref:`iteration hook <iterhook>`.

        Run every fonction hooked to this name (they should be iterators),
        and iterate over the chain of those iterators.

        :param str hookname: Name of the hook to apply.
        :param list arg: Positional arguments passed to the hooks.
        :param dict kwargs: Named arguments passed to the hooks.
        """
        for hook in sorted(self._hooks["iterhook"][hookname], key=str):
            yield from hook(*args, **kwargs)

    def get_plugin(self, keyword: str) -> Plugin:
        """Return the plugin with the given keyword.

        :raises NoMatch: If no (loaded) plugin was found with this keyword.
        """
        for plugin in self.values():
            if plugin.keyword == keyword:
                return plugin
        raise NoMatch(keyword, sorted(plugin.keyword for plugin in self.values()))

    def items(
        self, plugin_type: typing.Optional[str] = None
    ) -> typing.Iterable[typing.Tuple[str, Plugin]]:
        """Iterate over plugin keywords.

        :param typing.Optional[str] plugin_type: See :meth:`Loader.iter`.
        """
        for plugin in self.values(plugin_type):
            yield plugin.keyword

    def match(self, plugin_type: typing.Optional[str], value) -> Plugin:
        """Return the first plugin matching ``value``.

        A plugin ``Foo`` matches ``value`` if ``Foo.match(value)`` returns ``True``.

        :param typing.Optional[str] plugin_type: See :meth:`Loader.iter`.
        """
        for plugin in sorted(self.values(plugin_type), reverse=True):
            if plugin.match(value):
                return plugin
        raise NoMatch(value, sorted(self.iter(plugin_type)))

    def iter(
        self, plugin_type: typing.Union[str, None] = None
    ) -> typing.Iterable[Plugin]:
        """Iterate over keywords.

        :param typing.Optional[str] plugin_type: Type of the plugins to iterate over.

            - if ``None``, iterate over keywords of every (loaded) plugins;
            - else, iterate over keywords of plugins of this given type only.
        """
        if plugin_type is None:
            for ptype in self._plugins:
                yield from self._plugins[ptype]
        else:
            yield from self._plugins[plugin_type]



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Common utilities for readme renderers using Jinja2.

.. autoclass:: Jinja2ReadmeRenderer
   :members:

"""

# Can be removed starting with python3.11
from __future__ import annotations

import glob
import os
import pathlib
import typing

from ... import Plugin

if typing.TYPE_CHECKING:
    from ....tree import Tree


class Jinja2ReadmeRenderer(Plugin):
    """Default readme renderer using jinja2.

    This is an abstract class that defines a default README renderer for files.
    From within a template,
    the `macro <https://jinja.palletsprojects.com/en/3.0.x/templates/#macros>`__ ``render_readme``
    can be called to annotate a file, which:

    - looks for the first plugin that matches this file
      (that is, the first plugin where :meth:`Jinja2ReadmeRenderer.match` returns ``True``);
    - calls :meth:`Jinja2ReadmeRenderer.render`, and returns its return value.

    To implement such a renderer, in a subclass:

    - do one of:
        - set :attr:`Jinja2ReadmeRenderer.extensions` as a list of extensions:
          the README of any file ``foo`` is ``foo.{ext}``,
          the README of any directory is ``directory/README.{ext}``,
          where ``ext`` is one of the extensions listed here.
        - implement :meth:`Jinja2ReadmeRenderer.render`;
    - (optional) implement
      :meth:`Jinja2ReadmeRenderer.match` and :meth:`Jinja2ReadmeRenderer.get_readme`
      if the default implementation does not please you.
    """

    # pylint: disable=too-few-public-methods, abstract-method

    #: List of extensions of the READMEs (see :class:`Jinja2ReadmeRenderer`).
    extensions: typing.List[str] = []

    def match(self, tree: Tree) -> bool:
        """Return ``True`` if this plugin can handle the README of the argument."""
        # pylint: disable=arguments-renamed
        return self.get_readme(tree) is not None

    def get_readme(self, tree: Tree) -> typing.Optional[Tree]:
        """Return readme file for ``tree``, or ``None`` if there is no such README file."""
        if tree.is_dir():
            return self._get_readme_dir(tree)
        return self._get_readme_file(tree)

    def _get_readme_dir(self, tree):
        """Iterate over potential readme for the given directory ``tree``."""
        for ext in self.extensions:
            for filename in glob.iglob((tree.from_fs / f"*.{ext}").as_posix()):
                basename = os.path.basename(filename)
                if not tree.find(basename):
                    continue
                if basename.count(".") == 1:
                    if basename.split(".")[0].lower() == "readme":
                        return tree.root.find(
                            pathlib.Path(filename).relative_to(tree.root.from_fs)
                        )
        return None

    def _get_readme_file(self, tree):
        """Iterate over potential readme for the given file ``tree``."""
        for ext in self.extensions:
            for filename in glob.iglob(f"{tree.from_fs}.{ext}"):
                readme = tree.root.find(
                    pathlib.Path(filename).relative_to(tree.root.from_fs)
                )
                if readme:
                    return readme
        return None

    @staticmethod
    def render(tree: Tree) -> str:
        """Render argument as README.

        Return a string to be included when rendering the template.
        The functions and variables available in the template
        are described in :ref:`plugin_renderer_jinja2`.
        """
        # pylint: disable=unused-argument
        with open(tree.from_fs, encoding="utf8") as source:
            return source.read()



# Copyright Louis Paternault 2017-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Shell command: perform a (list of) shell command(s) on files."""

import glob
import io
import logging
import os
import pathlib
import re
import shlex
import subprocess
import tempfile
import threading

from ... import errors
from ...utils import yesno
from . import Action, Report

LOGGER = logging.getLogger("evariste")
STRACE_RE = re.compile(
    r'^\d* *open(at)?\((.*, )?"(?P<name>.*)",.*O_RDONLY.*\) = *[^ -].*'
)


class MissingOption(errors.EvaristeError):
    """No command was provided for action :class:`Command`."""

    def __init__(self, section, option, filename=None):
        super().__init__()
        self.filename = filename
        self.section = section
        self.option = option

    def __str__(self):
        # pylint: disable=line-too-long
        if self.filename is None:
            return f"Configuration for file is missing option '{self.option}' in section '{self.section}'."
        return f"Configuration for file '{self.filename}' is missing option '{self.option}' in section '{self.section}'."


def system_no_strace(command, path, log):
    """Run a system command.

    This function:
    - run command;
    - log standard output and error.
    """

    with subprocess.Popen(
        ["sh", "-c", command],
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        errors="replace",
        cwd=path.parent.from_fs.as_posix(),
    ) as process:

        log.write(process.communicate()[0])

        return process.returncode


def system_strace(command, path, log, depends):
    """Run a system command, analysing strace output to set `depends` files.

    This function:
    - run command;
    - log standard output and error;
    - track opened files.
    """

    def _process_strace_line(line):
        """Process output line of strace, and complete ``depends`` if relevant."""
        match = STRACE_RE.match(line)
        if match:
            name = pathlib.Path(path.parent.from_fs) / pathlib.Path(
                match.groupdict()["name"]
            )
            if name.resolve() != path.from_fs.resolve():  # pylint: disable=no-member
                if name in path.vcs:
                    depends.add(name)

    def _process_strace(pipe):
        """Process strace output, to find dependencies."""
        with open(pipe, mode="r", errors="replace", encoding="utf8") as file:
            for line in file:
                try:
                    _process_strace_line(line)
                except FileNotFoundError:
                    # File was deleted between the time its name is read and
                    # the time we check if it is a dependency: it can be
                    # discarded.
                    pass

    with tempfile.TemporaryDirectory() as tempdir:
        stdout = list(os.pipe())
        stderr = list(os.pipe())
        fifo = os.path.join(tempdir, str(id(path)))
        os.mkfifo(fifo)

        with subprocess.Popen(
            [
                "strace",
                "-f",
                "-o",
                fifo,
                "-e",
                "trace=open,openat",
                "sh",
                "-c",
                command,
            ],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            pass_fds=stdout + stderr,
            universal_newlines=True,
            errors="replace",
            cwd=path.parent.from_fs.as_posix(),
        ) as process:

            strace_thread = threading.Thread(
                target=_process_strace, daemon=True, kwargs={"pipe": fifo}
            )
            strace_thread.start()
            log.write(process.communicate()[0])
            os.unlink(fifo)
            strace_thread.join()
            for descriptor in stdout + stderr:
                os.close(descriptor)

    return process.returncode


class Command(Action):
    """System command"""

    keyword = "action.command"
    default_setup = {"strace": False}

    def get_targets(self, path):
        """Return the path of the targets."""
        if path.config[self.keyword]["targets"] is None:
            return []

        targets = []
        for item in shlex.split(path.config[self.keyword]["targets"]):
            for target in glob.glob(
                str(path.parent.from_fs.as_posix() / pathlib.Path(path.format(item)))
            ):
                # pylint: disable=duplicate-code
                targets.append(pathlib.Path(target).relative_to(path.root.from_fs))
        return targets

    def command(self, path):
        """Return the system command to run."""
        return path.config[self.keyword]["command"]

    def _run_command(self, path, command, *, log, depends):
        """Run the system command ``command``.

        - The list of dependencies is added to the set ``depends``.
        - The log is written at the end of ``log``.
        """
        LOGGER.info(f"Running command: {command}")

        if (yesno(path.config[self.keyword].get("strace", "false"))) or (
            "strace" not in path.config[self.keyword]
            and yesno(self.local.setup["strace"])
        ):
            returncode = system_strace(
                command=command,
                path=path,
                log=log,
                depends=depends,
            )
        else:
            returncode = system_no_strace(command=command, path=path, log=log)

        return returncode == 0

    def compile(self, path):
        command = path.format(self.command(path))
        depends = set()

        with io.StringIO() as log:
            log.write(f"$ {command}\n")

            success = self._run_command(path, command, log=log, depends=depends)

            return Report(
                path,
                success=success,
                targets=self.get_targets(path),
                depends=depends,
                log=log.getvalue(),
            )

    def match(self, value):
        # pylint: disable=unused-argument
        return False



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract class for jinja2 renderers.

See also :ref:`plugin_renderer_jinja2`.

.. autoclass:: Jinja2Renderer
   :members:

"""

# Can be removed starting with python3.11
from __future__ import annotations

import contextlib
import datetime
import functools
import os
import pathlib
import textwrap
import typing

import jinja2
import pkg_resources

from ....hooks import contexthook, iterhook
from ....utils import expand_path, smart_open
from ... import NoMatch, Plugin, errors, utils
from .file import Jinja2FileRenderer
from .readme import Jinja2ReadmeRenderer

if typing.TYPE_CHECKING:
    from ....builder import Builder
    from ....shared import Shared
    from ....tree import Tree

NOW = datetime.datetime.now()


class Jinja2Renderer(Plugin):
    """Abstract class for jinja2 renderers.

    To write your own renderer:

    - subclass this class;
    - define a default template name: :attr:`Jinja2Renderer.template`;
    - write such a template file,
      and place it in one of the :ref:`templatedirs <jinja2_options>`.
      The following template variables are defined
      and can be used in the template: :ref:`jinja2_template`;
    - you can also overwrite the methods defined here.

    You might also have a look at
    the implementation of the :class:`HTML renderer <evariste.plugins.renderer.html.HTMLRenderer>`.

    - Each file can be rendered in its own way:
      see :class:`~evariste.plugins.renderer.jinja2.file.Jinja2FileRenderer`
      (for instance, you might want to add a nice thumbnail to files that are images);
    - To define how files are annotated,
      see :class:`~evariste.plugins.renderer.jinja2.readme.Jinja2ReadmeRenderer`.
    """

    # pylint: disable=too-few-public-methods

    default_templatevar = {
        "date": NOW.strftime("%x"),
        "time": NOW.strftime("%X"),
        "datetime": NOW.strftime("%c"),
    }
    #: Name of the default template.
    template: str = None
    default_setup = {"destfile": "output"}

    def __init__(self, shared: Shared):
        super().__init__(shared)

        # Manage destination directory
        if self.local.setup["destdir"] is None:
            self.destdir = self.keyword
        else:
            self.destdir = utils.expand_path(self.local.setup["destdir"])
        try:
            os.makedirs(self.destdir, exist_ok=True)
        except FileExistsError as error:
            raise errors.EvaristeError(
                f"Cannot create directory '{self.destdir}'."
            ) from error

        # Create Jinja2 environment
        self.environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self._templatedirs()),
        )
        self.environment.filters["basename"] = os.path.basename
        self.environment.filters["yesno"] = utils.yesno

        # Dictionary of README files and renderers
        self.readmes = {}

    @iterhook("Tree.prune_before")
    def get_readme(self, tree: Tree) -> Tree:
        """Iterate the only README file for ``tree``.

        If there is such a README file, iterate over it (a single value);
        otherwise, iterate nothing.

        Side effect:
        Store a (partial) function in ``self.readmes[tree.from_source]`` to render this README file.
        """
        for plugin_type in self.iter_subplugins("readme"):
            try:
                renderer = self.shared.builder.plugins.match(plugin_type, tree)
                readme = renderer.get_readme(tree)
                self.readmes[tree.from_source] = functools.partial(
                    renderer.render, readme
                )
                yield readme.from_source
                return
            except NoMatch:
                continue
        return

    @utils.cached_iterator
    def iter_subplugins(self, subtype: str) -> typing.Iterable[Plugin]:
        """Iterate over subplugins of type ``subtype``.

        This method iterates plugins (as their keywords)
        ``{keyword}.{subtype}``, where ``keyword`` is the attribute of this class,
        or its subclasses.

        For instance, given that:

        - the correct plugins are loaded;
        - plugin :class:`renderer.html <evariste.plugins.renderer.html.HTMLRenderer>`
          is a subclass of :class:`renderer.jinja2 <Jinja2Renderer>`,

        call to ``Jinja2Renderer.iter_subplugins(HtmlRenderer(), "readme")``
        will yield: ``renderer.html.readme``, ``renderer.html.readme.mdwn``…
        """
        for parent in self.__class__.mro():  # pylint: disable=no-member
            if not hasattr(parent, "keyword"):
                break
            if parent.keyword is None:
                continue
            yield f"{parent.keyword}.{subtype}"

    def _templatedirs(self):
        """Iterator over the directories in which templates may exist.

        - Directories are returned as strings;
        - directories may not exist.
        """
        if self.local.setup["templatedirs"] is not None:
            yield from utils.expand_path(self.local.setup["templatedirs"]).split()
        yield pkg_resources.resource_filename(  # pylint: disable=no-member
            self.__class__.__module__, os.path.join("data", "templates")
        )
        yield from [
            os.path.join(utils.expand_path(path), "templates")
            for path in [
                ".evariste",
                "~/.config/evariste",
                "~/.evariste",
                "/usr/share/evariste",
            ]
        ]

    def render_tree(self, tree: Tree) -> str:
        """Render the tree using templates, and return the string."""
        # Copy targets to destination
        for file in tree.walk(dirs=False, files=True):
            if file.report.success:
                for target in file.report.targets:
                    utils.copy(
                        (file.root.from_fs / target).as_posix(),
                        (pathlib.Path(self.destdir) / target).as_posix(),
                    )
        # Select main template
        if self.local.setup["template"] is None:
            template = self.template
        else:
            template = self.local.setup["template"]

        # Create template loading file renderers
        content = ""
        for plugin_type in self.iter_subplugins("file"):
            for subrenderer in self.shared.builder.plugins.values(plugin_type):
                if subrenderer.extension is None:
                    subtemplate = subrenderer.template
                else:
                    subtemplate = f"{subrenderer.template}.{subrenderer.extension}"
                content += textwrap.dedent(
                    f"""\
                        {{%
                            from "file/{subtemplate}"
                            import file as file_{subrenderer.template}
                            with context
                        %}}
                        """
                )
        content += f"""{{% include "{template}" %}}"""

        # Render template
        return self.environment.from_string(content).render(
            {
                "destdir": pathlib.Path(self.destdir),
                "shared": self.shared,
                "local": self.local,
                "sourcepath": self._sourcepath,
                "render_file": self._render_file,
                "render_readme": self._render_readme,
                "render_template": self._render_template,
                "templatevar": self._get_templatevar(),
                "tree": tree,
            }
        )

    @contexthook("Builder.compile")
    @contextlib.contextmanager
    def render(self, builder: Builder) -> None:
        """Render the tree as a file, and write result into the destination file."""
        yield

        with smart_open(expand_path(self.local.setup["destfile"]), "w") as destfile:
            destfile.write(self.render_tree(builder.tree))

    def _get_templatevar(self):
        """Return the template variables.

        - First, update it with the default variables of this class
          (`self.default_templatevar`), then its ancestors.
        - Then, update it with the variables defined in the setup file.
        """
        templatevar = {}
        for parent in reversed(self.__class__.mro()):  # pylint: disable=no-member
            templatevar.update(getattr(parent, "default_templatevar", {}))
        templatevar.update(self.shared.setup[f"{self.keyword}.templatevar"])
        return templatevar

    @jinja2.pass_context
    def _render_file(self, context, filename):
        """Render ``context['file']``, which is a :class:`pathlib.Path`."""
        for plugin_type in self.iter_subplugins("file"):
            try:
                return self.shared.builder.plugins.match(plugin_type, filename).render(
                    filename, context
                )
            except NoMatch:
                continue
        return ""

    @jinja2.pass_context
    def _sourcepath(self, context, tree):
        """Return the path to the source file or archive.

        This functions builds the archive before returning its path. It can be
        called several times: the archive will be built only once.
        """
        # pylint: disable=no-self-use
        return tree.make_archive(context["destdir"])

    @jinja2.pass_context
    def _render_readme(self, context, tree):
        """Return the code for the readme of `tree`."""
        # pylint: disable=unused-argument
        if tree.from_source in self.readmes:
            return self.readmes[tree.from_source]()
        return ""

    @jinja2.pass_context
    def _render_template(self, context, template):
        """Render template given in argument."""
        return textwrap.indent(
            self.environment.get_or_select_template(template).render(context), "  "
        )



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin to evaluate and store modification times of files.

It is used not to re-compile files that have not changed.
"""

import contextlib
import functools
import os
import pathlib
from datetime import datetime

from .. import plugins
from ..hooks import contexthook, methodhook
from ..tree import File


class Changed(plugins.Plugin):
    """Evaluate and store modification times of files."""

    keyword = "changed"
    default_setup = {"time": "vcs"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.files_to_compile = set()
        self.files_to_ignore = set()

    def compile(self, tree):
        """Return `True` iff file `tree` is to be compiled.

        It is not compiled if:
        - nothing has changed since last compilation;
        - it is a dependency of a file described in previous item.
        """
        return (tree.from_fs.resolve() in self.files_to_compile) and (
            tree.from_fs.resolve() not in self.files_to_ignore
        )

    # Hooks
    @methodhook("File.make_archive")
    def make_archive(self, function):
        """Save archive path."""

        @functools.wraps(function)
        def wrapped(tree, destdir):
            """Wrapped function."""
            # pylint: disable=unused-argument
            value = function(tree, destdir)
            if self.compile(tree):
                self.local.tree[tree.from_source]["archivepath"] = value
            return value

        return wrapped

    @contexthook("Tree")
    @contextlib.contextmanager
    def build_tree_item(self, tree, *args, **kwargs):  # pylint: disable=unused-argument
        """Compare data with cache to see if a compilation is needed."""

        yield

        if not isinstance(tree, File):
            return

        # If the --always-compile argument was given to the command line call,
        # do compile.
        if self.shared.setup["arguments"]["always_compile"]:
            self.files_to_compile.add(tree.from_fs.resolve())
            return

        # If tree is not in hash, compile it
        if (self.local.tree[tree.from_source] is None) or (
            self.local.tree[tree.from_source] == {}
        ):
            self.files_to_compile.add(tree.from_fs.resolve())
            return

        # Otherwise, check if data hash has changed
        for key in ["depends", "targets"]:
            dependencies = self.local.tree[tree.from_source][key]
            for depend in dependencies:
                cache_mtime = dependencies[depend][self.local.setup["time"]]
                if self.local.setup["time"] == "vcs":
                    file_mtime = tree.vcs.last_modified(pathlib.Path(depend))
                else:
                    try:
                        file_mtime = datetime.fromtimestamp(os.path.getmtime(depend))
                    except FileNotFoundError:
                        self.files_to_compile.add(tree.from_fs.resolve())
                        return
                if cache_mtime != file_mtime:
                    self.files_to_compile.add(tree.from_fs.resolve())
                    return

    @contexthook("File.compile")
    @contextlib.contextmanager
    def file_compile(self, tree):
        """Store the modification time of file and its dependencies."""

        yield

        if not isinstance(tree, File):
            return

        if not tree.report.success:
            self.files_to_compile.discard(tree.from_fs.resolve())
            return

        if self.local.tree[tree.from_source] is None:
            self.local.tree[tree.from_source] = {}

        # Saving file and dependencies data into cache
        self.local.tree[tree.from_source]["depends"] = {
            file.as_posix(): {
                "vcs": tree.vcs.last_modified(file),
                "fs": datetime.fromtimestamp(os.path.getmtime(file.as_posix())),
            }
            for file in tree.full_depends()
        }

        self.local.tree[tree.from_source]["targets"] = {}
        for target in tree.report.targets:
            targetpath = tree.root.from_fs / target
            self.local.tree[tree.from_source]["targets"][targetpath.as_posix()] = {
                "vcs": tree.vcs.last_modified(targetpath),
                "fs": datetime.fromtimestamp(os.path.getmtime(targetpath.as_posix())),
            }

        # Mark dependencies as ignored (not to be compiled)
        self.files_to_ignore.update([depend.resolve() for depend in tree.depends()])
